/*
** Copyright (c) 1995-2002
**      Daniel O'Connor and Genesis Software.  All rights reserved.
**
** Originally written by Michael Smith.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions
** are met:
** 1. Redistributions of source code must retain the above copyright
**    notice, this list of conditions and the following disclaimer as
**    the first lines of this file unmodified.
** 2. Redistributions in binary form must reproduce the above copyright
**    notice, this list of conditions and the following disclaimer in the
**    documentation and/or other materials provided with the distribution.
** 3. The name of the author may not be used to endorse or promote products
**    derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY Daniel O'Connor AND Genesis Software ``AS IS'' 
** AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
** IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
** ARE DISCLAIMED. IN NO EVENT SHALL Michael Smith AND OR Genesis Software BE
** LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
** CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
** SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
** INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
** CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
** ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
** POSSIBILITY OF SUCH DAMAGE.
**
**      $Id: mdsio.c,v 1.22 2002/03/28 01:22:53 doconnor Exp $
*/
/*
** RS-485 driver for 8250-family UARTs.
**
** Assorted wisdom, comments and code fragments taken from the FreeBSD 'sio' 
** driver; thanks and credit to all who have worked on it.
**
** See mdsio(4) for a description of the flags values and the programming
** interface.
*/

#include <sys/types.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/filio.h>
#include <sys/conf.h>
#include <sys/kernel.h>
#include <sys/malloc.h>
#include <sys/vnode.h>

#include <sys/select.h>

#include <sys/poll.h>

#include <machine/clock.h>

#include <machine/bus_pio.h> /* Newbus */
#include <machine/bus_memio.h>
#include <machine/bus.h>
#include <machine/resource.h>
#include <sys/bus.h>
#include <sys/rman.h>

#include <sys/devicestat.h> /* For devstat */

#include <pci/pcireg.h>
#include <pci/pcivar.h>

#include <isa/isareg.h>
#include <isa/isavar.h>

#include <vm/vm.h>
#include <vm/vm_param.h>
#include <vm/pmap.h>

#include <sys/sysctl.h>

#include <sys/tty.h>
#include <i386/isa/isa.h>
#include <i386/isa/isa_device.h>
#include <isa/sioreg.h>
#if __FreeBSD_version < 450000
#include <isa/ic/ns16550.h>
#else
#include <dev/ic/ns16550.h>
#endif
#ifdef EXTREE
#include <mdsio.h>
#include <mdsioreg.h>
#include <mdsiovar.h>
#else
#include <sys/mdsio.h>
#include <dev/mdsio/mdsioreg.h>
#include <dev/mdsio/mdsiovar.h>
#endif

/*
** These macros require that all routines have a pointer 'md' pointing
** to the softc struct for the current port.
*/
#ifdef MDSIODEBUG
#define debug(fmt, args...)	printf("mdsio%d(%s): " fmt "\n" ,md->unit ,  __FUNCTION__ , ##args)
#define debug2(fmt, args...)	printf("mdsio(%s): " fmt "\n",  __FUNCTION__ , ##args)
#else
#define debug(fmt, args...)	
#define debug2(fmt, args...)	
#endif
#define error(fmt, args...)	printf("mdsio%d: " fmt "\n" , md->unit , ##args);

/*******************************************************************************
** Local prototypes
*/

static int		mdsioprobe(device_t dev, struct __md_res *res, int clk);
static int		mdsioprobe_isa(device_t dev);
static int		mdsioprobe_pci(device_t dev);
static int		mdsioattach(device_t dev);
static int		mdsioattach_isa(device_t dev);
static int		mdsioattach_pci(device_t dev);
#ifdef MSIO_USE_IDENTIFY
static void     	mdsioidentify_isa(driver_t *, device_t);  
#endif
static int		mdsiodetach(device_t dev);

static timeout_t	mdsiocomplain;
static timeout_t	mdsiotimeout;

static void		mdsiointr(void *arg);

static int		combrd(int clk, int baud);
static int		sysctl_hw_mdsio_txcontrol(SYSCTL_HANDLER_ARGS);
static void		mdsio_update_options(device_t dev);

/*******************************************************************************
** ISA Newbus glue
*/

static device_method_t mdsio_methods_isa[] = {
    /* Device Interface */
    DEVMETHOD(device_probe,	mdsioprobe_isa),
    DEVMETHOD(device_attach,	mdsioattach_isa),
#ifdef MDSIO_USE_IDENTIFY
    DEVMETHOD(device_identify,	mdsioidentify_isa),
#endif
    DEVMETHOD(device_detach,	mdsiodetach),

    /* Bus Interface */
    DEVMETHOD(bus_print_child,	bus_generic_print_child),
    DEVMETHOD(bus_driver_added,	bus_generic_driver_added),

    {0, 0}
};

static driver_t mdsio_driver_isa = {
    DEVNAME,
    mdsio_methods_isa,
    sizeof(struct mdsio_softc)
};

DRIVER_MODULE(mdsio, isa, mdsio_driver_isa, mdsio_devclass, 0, 0);

/*******************************************************************************
** PCI Newbus glue
*/

static device_method_t mdsio_methods_pci[] = {
    /* Device Interface */
    DEVMETHOD(device_probe,	mdsioprobe_pci),
    DEVMETHOD(device_attach,	mdsioattach_pci),
    DEVMETHOD(device_detach,	mdsiodetach),

    /* Bus Interface */
    DEVMETHOD(bus_print_child,	bus_generic_print_child),
    DEVMETHOD(bus_driver_added,	bus_generic_driver_added),

    {0, 0}
};

static driver_t mdsio_driver_pci = {
    DEVNAME,
    mdsio_methods_pci,
    sizeof(struct mdsio_softc)
};

DRIVER_MODULE(mdsio, pci, mdsio_driver_pci, mdsio_devclass, 0, 0);

/*******************************************************************************
** cdevsw stuff
**
*/
static d_open_t		mdsioopen;
static d_close_t	mdsioclose;
static d_read_t		mdsioread;
static d_write_t	mdsiowrite;
static d_ioctl_t	mdsioioctl;
static d_poll_t		mdsiopoll;

static struct cdevsw mdsio_cdevsw = {
    mdsioopen,	/* open */
    mdsioclose,	/* close */
    mdsioread,	/* read */
    mdsiowrite,	/* write */
    mdsioioctl,	/* ioctl */
    mdsiopoll,	/* poll */
    nommap,	/* mmap */
    nostrategy,	/* strategy */
    DEVNAME,	/* name */
    CDEV_MAJOR,	/* maj */
    nodump,	/* dump */
    nopsize,	/* psize */
    0,		/* flags */
    -1		/* bmaj */
};

/*******************************************************************************
** Device driver functions start here
*/

#ifdef MDSIO_USE_IDENTIFY
/*
** mdsioidentify_isa
**
** Lie and say we find a MDSIO card at a fixed address so we can be kldloaded
*/

#define MDSIO_FAKEBASE 0x3e8 /* COM3 */
#define MDSIO_FAKEIRQ 10

static void
mdsioidentify_isa(driver_t *driver, device_t parent) {
    device_t        child;

    debug2("mdsio: saying base = 0x%x, IRQ = %d", MDSIO_FAKEBASE, MDSIO_FAKEIRQ);
    
    child = BUS_ADD_CHILD(parent, ISA_ORDER_SPECULATIVE, "mdsio", -1);
    device_set_desc_copy(child, "PCL-740 RS-232/485/Current-Loop Interface Card");
    device_set_driver(child, driver);
    bus_set_resource(child, SYS_RES_IRQ, 0, MDSIO_FAKEIRQ, 1);
    bus_set_resource(child, SYS_RES_IOPORT, 0, MDSIO_FAKEBASE, IO_COMSIZE);
}
#endif

/*
** mdsioprobe_isa
**
** Allocate the port to probe the device and then call the generic probe
*/
static int
mdsioprobe_isa(device_t dev) {
    struct __md_res	res;
    int			result;
    
    res.rid = 0;
    res.reshandle = bus_alloc_resource(dev, SYS_RES_IOPORT, &res.rid,
				  0, ~0, IO_COMSIZE, RF_ACTIVE);
    if (!res.reshandle) {
	debug2("isa - bus_alloc_resource for port returned NULL");
	return ENXIO;
    }
    res.sc_st = rman_get_bustag(res.reshandle);
    res.sc_sh = rman_get_bushandle(res.reshandle);
    
    debug2("mdsio: isa - base address 0x%08lx", rman_get_start(res.reshandle));

    result = mdsioprobe(dev, &res, 1843200);

    bus_release_resource(dev, SYS_RES_IOPORT, res.rid, res.reshandle);

    return(result);
}

/*
** mdsioprobe_pci
**
** Allocate the port to probe the device and then call the generic probe
*/
static int
mdsioprobe_pci(device_t dev) {
    struct __md_res	res;
    int			result;
    
    if ((pci_get_device(dev) == PCI_DEV_ID) &&
	(pci_get_vendor(dev) == PCI_VND_ID)) {
	debug2("Found PCI device ID match");
	device_set_desc(dev, "Decision PCCOM RS232/422/486 adapter");
    } else
	return (ENXIO);

    res.rid = PLX_PCIBAR3;
    res.reshandle = bus_alloc_resource(dev, SYS_RES_IOPORT, &res.rid,
				  0, ~0, 1, RF_ACTIVE);
    if (!res.reshandle) {
	debug2("pci - bus_alloc_resource for port returned NULL");
	return ENXIO;
    }
    res.sc_st = rman_get_bustag(res.reshandle);
    res.sc_sh = rman_get_bushandle(res.reshandle);
    
    debug2("mdsio: pci - base address 0x%08lx", rman_get_start(res.reshandle));

    result = mdsioprobe(dev, &res, 18432000);

    bus_release_resource(dev, SYS_RES_IOPORT, res.rid, res.reshandle);

    debug2("mdsio: pci - done");
    return(result);
}

/*
** mdsioprobe
**
** Renamed and recontextified sioprobe(); find and test the UART.
*/
static int
mdsioprobe(device_t dev, struct __md_res *res, int clk) {
    u_char		mcr_image;
    u_char		failures[10] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
    u_int		irqmap[4];
    u_int		irqs;
    int			result;
    int			fn;
    u_long		xirq;

    debug2("mdsio: being probed..");
    
    mcr_image = MCR_IENABLE;    	/* prepart to set MCR_IENABLE */
  
    bzero(failures, sizeof failures);
    /*
     * We don't want to get actual interrupts, just masked ones.
     * Interrupts from this line should already be masked in the ICU,
     * but mask them in the processor as well in case there are some
     * (misconfigured) shared interrupts.
     */
    debug2("mdsio: disabling interrupts");
    disable_intr();

    /*
     * Initialize the speed and the word size and wait long enough to
     * drain the maximum of 16 bytes of junk in device output queues.
     * The speed is undefined after a master reset and must be set
     * before relying on anything related to output.  There may be
     * junk after a (very fast) soft reboot and (apparently) after
     * master reset.
     */
    debug2("mdsio: Initialise");
    m_outb(*res, com_lctl, CFCR_DLAB);
    m_outb(*res, com_dlbl, combrd(clk, 9600) & 0xff);
    m_outb(*res, com_dlbh, (u_int) combrd(clk, 9600) >> 8);
    m_outb(*res, com_lctl, CFCR_8BITS);
    DELAY((16 + 1) * 1000000 / (9600 / 10));

    /*
     * Enable the interrupt gate and disable device interupts.  This
     * should leave the device driving the interrupt line low and
     * guarantee an edge trigger if an interrupt can be generated.
     */
    debug2("mdsio: seeing if we got the right interrupt");
    m_outb(*res, com_mcr, mcr_image);
    m_outb(*res, com_ier, 0);
    DELAY(1000);		/* XXX */
    irqmap[0] = isa_irq_pending();
    debug2("mdsio: irqmap[0] = 0x%x", irqmap[0]);

    /*
     * Attempt to set loopback mode so that we can send a null byte
     * without annoying any external device.
     */
    debug2("mdsio: setting loopback");
    m_outb(*res, com_mcr, mcr_image | MCR_LOOPBACK);
    
    /*
     * Attempt to generate an output interrupt.  On 8250's, setting
     * IER_ETXRDY generates an interrupt independent of the current
     * setting and independent of whether the THR is empty.  On 16450's,
     * setting IER_ETXRDY generates an interrupt independent of the
     * current setting.  On 16550A's, setting IER_ETXRDY only
     * generates an interrupt when IER_ETXRDY is not already set.
     */
    m_outb(*res, com_ier, IER_ETXRDY);

    /*
     * On some 16x50 incompatibles, setting IER_ETXRDY doesn't generate
     * an interrupt.  They'd better generate one for actually doing
     * output.  Loopback may be broken on the same incompatibles but
     * it's unlikely to do more than allow the null byte out.
     */
    m_outb(*res, com_data, 0);
    DELAY((1 + 2) * 1000000 / (9600 / 10));

    /*
     * Turn off loopback mode so that the interrupt gate works again
     * (MCR_IENABLE was hidden).  This should leave the device driving
     * an interrupt line high.  It doesn't matter if the interrupt
     * line oscillates while we are not looking at it, since interrupts
     * are disabled.
     */
    m_outb(*res, com_mcr, mcr_image);

    /*
     * Check that
     *      o the CFCR, IER and MCR in UART hold the values written to them
     *        (the values happen to be all distinct - this is good for
     *        avoiding false positive tests from bus echoes).
     *      o an output interrupt is generated and its vector is correct.
     *      o the interrupt goes away when the IIR in the UART is read.
     */
    failures[0] = m_inb(*res, com_lctl) - CFCR_8BITS;
    failures[1] = m_inb(*res, com_ier) - IER_ETXRDY;
    failures[2] = m_inb(*res, com_mcr) - mcr_image;
    DELAY(10000);
    irqmap[1] = isa_irq_pending();
    failures[4] = (m_inb(*res, com_iir) & IIR_IMASK) - IIR_TXRDY;
    DELAY(1000);
    irqmap[2] = isa_irq_pending();
    failures[6] = (m_inb(*res, com_iir) & IIR_IMASK) - IIR_NOPEND;
    debug2("mdsio: irqmap[1] = 0x%0x, irqmap[2] = 0x%x", irqmap[1], irqmap[2]);

    /*
     * Turn off all device interrupts and check that they go off properly.
     * Leave MCR_IENABLE alone, it gates the OUT2 output of the UART to
     * the ICU input.  Closing the gate would give a floating ICU input
     * (unless there is another device driving it) and spurious interrupts.
     * (On the system that this was first tested on, the input floats high
     * and gives a (masked) interrupt as soon as the gate is closed.)
     */
    debug2("mdsio: turning off interrupts");
    m_outb(*res, com_ier, 0);
    m_outb(*res, com_lctl, CFCR_8BITS);    /* dummy to avoid bus echo */
    failures[7] = m_inb(*res, com_ier);
    DELAY(1000);
    irqmap[3] = isa_irq_pending();
    failures[9] = (m_inb(*res, com_iir) & IIR_IMASK) - IIR_NOPEND;
    debug2("mdsio: irqmap[3] = 0x%0x, failures[9] = 0x%x", irqmap[3], failures[9]);

    debug2("mdsio: enabling interrupts");
    enable_intr();

    irqs = irqmap[1] & ~irqmap[0];

    if (bus_get_resource(dev, SYS_RES_IRQ, 0, &xirq, NULL) == 0 &&
	((1 << xirq) & irqs) == 0) {
	device_printf(dev, "configured irq %ld not in bitmap of probed irqs %#x\n", xirq,
		      irqs);
	device_printf(dev, "irq maps: %#x %#x %#x %#x\n",
		      irqmap[0], irqmap[1], irqmap[2], irqmap[3]);
    }
    
    result = 0;
    for (fn = 0; fn < sizeof failures; ++fn) {
	if (failures[fn]) {
	    m_outb(*res, com_mcr, 0);
	    result = ENXIO;
	    printf("mdsio%d: probe failed test(s):",
		   device_get_unit(dev));
	    for (fn = 0; fn < sizeof failures; ++fn)
		if (failures[fn])
		    printf(" %d", fn);
	    printf("\n");
	    break;
	}
    }
    return (result);
}

/*
** mdsioattach_isa
**
** ISA specific attach routine
*/

static int
mdsioattach_isa(device_t dev) {
    struct mdsio_softc	*md = device_get_softc(dev);
    int			error = 0;

    md->iobase.rid = 0;
    md->iobase.reshandle = bus_alloc_resource(dev, SYS_RES_IOPORT, &md->iobase.rid,
				      0, ~0, IO_COMSIZE, RF_ACTIVE);
    md->iobase.sc_st = rman_get_bustag(md->iobase.reshandle);
    md->iobase.sc_sh = rman_get_bushandle(md->iobase.reshandle);

    if (!md->iobase.reshandle) {
	device_printf(dev, "isa - couldn't allocate IOPORT\n");
	error = ENXIO;
	goto fail;
    }

    /* 1.8432 Mhz clock */
    md->clk = 1843200;
    
    /* Call generic attach */
    if (mdsioattach(dev)) {
	error = ENXIO;
	goto fail;
    }

    debug("allocating IRQ resource");
    
    md->irqrid = 0;
    md->irqres = bus_alloc_resource(dev, SYS_RES_IRQ, &md->irqrid, 0ul, ~0ul, 1,
				    RF_ACTIVE);
    debug("isa - irqrid is 0x%x", md->irqrid);
    
    if (md->irqres) {
	debug("isa - allocated IRQ");

	if (bus_setup_intr(dev, md->irqres, INTR_TYPE_TTY | INTR_TYPE_FAST,
			   mdsiointr, md, &md->irq_ih)) {

	    device_printf(dev, "isa - couldn't activate interrupt\n");
	    error = ENXIO;
	    goto fail;
	}
    } else {
	device_printf(dev, "isa - bus_alloc_resource returned NULL!\n");
	error = ENXIO;
	goto fail;
    }
    
    /* Add devstat handle */
    if ((md->md_ds = malloc(sizeof(struct devstat), M_DEVBUF, M_NOWAIT)) == NULL) {
	device_printf(dev, "isa - Couldn't allocate devstat handle\n");
	goto fail;
    }

    return(0);

 fail:
    if (md->irqres != NULL)
	bus_release_resource(dev, SYS_RES_IRQ, 0, md->irqres);
    
    if (md->iobase.reshandle != NULL)
	bus_release_resource(dev, SYS_RES_IOPORT, md->iobase.rid, md->iobase.reshandle);

    if (md->md_ds != NULL)
	free(md->md_ds, M_DEVBUF);
    
    return(error);
}

/*
** mdsioattach_isa
**
** PCI specific attach routine
*/

static int
mdsioattach_pci(device_t dev) {
    struct mdsio_softc	*md = device_get_softc(dev);
    int			error = 0;

    md->iobase.rid = PLX_PCIBAR3;
    md->iobase.reshandle = bus_alloc_resource(dev, SYS_RES_IOPORT, &md->iobase.rid,
				      0, ~0, IO_COMSIZE, RF_ACTIVE);
    md->iobase.sc_st = rman_get_bustag(md->iobase.reshandle);
    md->iobase.sc_sh = rman_get_bushandle(md->iobase.reshandle);

    if (!md->iobase.reshandle) {
	device_printf(dev, "pci - couldn't allocate IOPORT\n");
	error = ENXIO;
	goto fail;
    }

    /* 1.84320 Mhz clock */
    md->clk = 1843200;

    /* Enable RTS TX control */
    device_set_flags(dev, device_get_flags(dev) | 0x24);

    /* Call generic attach */
    if (mdsioattach(dev)) {
	error = ENXIO;
	goto fail;
    }

    debug("pci - allocating IRQ resource");
    
    md->irqrid = 0;
    md->irqres = bus_alloc_resource(dev, SYS_RES_IRQ, &md->irqrid, 0ul, ~0ul, 1,
				    RF_ACTIVE);
    debug("pci - irqrid is 0x%x", md->irqrid);
    
    if (!md->irqres) {
	debug("pci - failed to allocate dedicated IRQ");
	md->irqres = bus_alloc_resource(dev, SYS_RES_IRQ, &md->irqrid, 0ul, ~0ul, 1,
				    RF_SHAREABLE | RF_ACTIVE);
	debug("pci - irqrid is 0x%x", md->irqrid);
    }
    
    if (md->irqres) {
	debug("pci - allocated IRQ");

	if (bus_setup_intr(dev, md->irqres, INTR_TYPE_TTY, mdsiointr, md, &md->irq_ih)) {
	    device_printf(dev, "pci - couldn't activate interrupt\n");
	    error = ENXIO;
	    goto fail;
	}
    } else {
	device_printf(dev, "pci - bus_alloc_resource returned NULL!\n");
	error = ENXIO;
	goto fail;
    }
    
    return(0);

 fail:
    if (md->irqres != NULL)
	bus_release_resource(dev, SYS_RES_IRQ, 0, md->irqres);
    
    if (md->iobase.reshandle != NULL)
	bus_release_resource(dev, SYS_RES_IOPORT, md->iobase.rid, md->iobase.reshandle);

    if (md->md_ds != NULL)
	free(md->md_ds, M_DEVBUF);
    
    return(error);
}

/*
** mdsioattach
**
** Set ourselves up to talk to the port.  Again derived from the
** corresponding sio routine.
*/

static int
mdsioattach(device_t dev) {
    struct mdsio_softc *md = device_get_softc(dev);
    u_char 		scr;
    u_char  		scr1;
    u_char  		scr2;
    
    md->unit = device_get_unit(dev);		/* save for reverse lookup */
    md->ints = md->errors = md->rxbytes = 0;	/* zero stats */
    
    /*
     * mdsioprobe() has initialized the device registers as follows:
     *      o cfcr = CFCR_8BITS.
     *        It is most important that CFCR_DLAB is off, so that the
     *        data port is not hidden when we enable interrupts.
     *      o ier = 0.
     *        Interrupts are only enabled when the line is open.
     *      o mcr = MCR_IENABLE
     */

    md->rxhead = md->rxtail = 0;	/* initialise the ringbuffer */
    md->hasfifo = FALSE;		/* at least, so we assume */

    /* attempt to determine UART type */
    device_printf(dev, "type");
   
    scr = m_inb(md->iobase, com_scr);
    m_outb(md->iobase, com_scr, 0xa5);
    scr1 = m_inb(md->iobase, com_scr);
    m_outb(md->iobase, com_scr, 0x5a);
    scr2 = m_inb(md->iobase, com_scr);
    m_outb(md->iobase, com_scr, scr);
    if (scr1 != 0xa5 || scr2 != 0x5a) {
	printf(" 8250");
    } else {
	m_outb(md->iobase, com_fifo, FIFO_ENABLE | FIFO_RX_HIGH);
	DELAY(1000);
	switch (m_inb(md->iobase, com_iir) & IIR_FIFO_MASK) {
	case FIFO_RX_LOW:
	    printf(" 16450");
	    break;
	case FIFO_RX_MEDL:
	    printf(" 16450?");
	    break;
	case FIFO_RX_MEDH:
	    printf(" 16550?");
	    break;
	case FIFO_RX_HIGH:
	     printf(" 16550A");
	    md->hasfifo = TRUE;

	    break;
	}
	m_outb(md->iobase, com_fifo, 0);
    }

    printf("\n");
    mdsio_update_options(dev);		/* Setup the options based on the flag settings */

    debug("Clock rate %ld Hz", md->clk);
    debug("making device node");
    
    md->dev_handle = make_dev(&mdsio_cdevsw, md->unit, 0, 0, 0644, DEVNAME);
    
    /* Add devstat handle */
    if ((md->md_ds = malloc(sizeof(struct devstat), M_DEVBUF, M_NOWAIT)) == NULL) {
	device_printf(dev, "Couldn't allocate devstat handle\n");
	return (-1);
    }
	
    bzero(md->md_ds, sizeof(struct devstat));

    devstat_add_entry(md->md_ds, "mdsio", md->unit, 0,
		      DEVSTAT_NO_BLOCKSIZE | DEVSTAT_NO_ORDERED_TAGS,
		      DEVSTAT_TYPE_DIRECT | DEVSTAT_TYPE_IF_OTHER,
		      DEVSTAT_PRIORITY_OTHER);

    /*
     * Create the sysctl tree
     */
    sysctl_ctx_init(&md->sysctl_ctx);
    md->sysctl_tree = SYSCTL_ADD_NODE(&md->sysctl_ctx,
				      SYSCTL_STATIC_CHILDREN(_hw), OID_AUTO,
				      device_get_nameunit(dev), CTLFLAG_RD, 0, "");
    if (md->sysctl_tree == NULL)
	return(-1);

    SYSCTL_ADD_PROC(&md->sysctl_ctx, SYSCTL_CHILDREN(md->sysctl_tree),
		    OID_AUTO, "txcontrol", CTLTYPE_INT | CTLFLAG_RW | CTLFLAG_PRISON,
		    dev, 0, &sysctl_hw_mdsio_txcontrol, "I", "MDSIO transmit control method");
    return(0);
}

/*
** mdsiodetach
** 
** Detach the device if not busy
**
*/
static int
mdsiodetach(device_t dev) {
    struct mdsio_softc	*md = device_get_softc(dev);
    int			s;

    /* Don't really want an interrupt in here :) */
    s = spltty();
    
    if(md->state & MD_OPEN) {
	device_printf(dev, "In use, not leaving\n");
	return(EBUSY);
    }

    device_printf(dev, "leaving...\n");
    
    /* Remove devstat handle */
    devstat_remove_entry(md->md_ds);
    free(md->md_ds, M_DEVBUF);
    
    /* Unregister ourselves */
    destroy_dev(md->dev_handle);

    /* Free our bus resources */
    bus_generic_detach(dev);

    bus_release_resource(dev, SYS_RES_IOPORT, md->iobase.rid, md->iobase.reshandle);

    bus_teardown_intr(dev, md->irqres, md->irq_ih);
    bus_release_resource(dev, SYS_RES_IRQ, md->irqrid, md->irqres);

    sysctl_ctx_free(&md->sysctl_ctx);

    splx(s);

    return(0);
}

/*
** mdsioopen
**
** Default the UART configuration and enable receiver interrupts.
** Only one opener is allowed at a time. 
*/

int
mdsioopen(dev_t dev, int flag, int fmt, struct proc *p) {
    int			unit = minor(dev);
    struct mdsio_softc	*md = devclass_get_softc(mdsio_devclass, unit);
    int			s;
    
    if (md->state & MD_OPEN)		/* already open */
	return(EBUSY);

    s = spltty();
    if (md->hasfifo) {
	/*
	 * (Re)enable and drain fifos.
	 *
	 * Certain SMC chips cause problems if the fifos are enabled
	 * while input is ready.  Turn off the fifo if necessary to
	 * clear the input.  We test the input ready bit after
	 * enabling the fifos to handle races between enabling and
	 * fresh input.  
	 */
	while (TRUE) {
	    m_outb(md->iobase, com_fifo, FIFO_RCV_RST | FIFO_XMT_RST | FIFO_ENABLE | MDSIO_FTL);
	    DELAY(100);
	    if (!(m_inb(md->iobase, com_lsr) & LSR_RXRDY))
		break;
	    m_outb(md->iobase, com_fifo, 0);
	    DELAY(100);
	    (void) m_inb(md->iobase, com_data);
	}
    }
    md->rxtail = md->rxhead = 0;		/* empty ringbuffers */
    md->txtail = md->txhead = 0;
    md->eflag = 0;				/* no errors */
    md->rsel.si_flags = 0;			/* nothing selecting */
    md->rsel.si_pid = 0;
    md->wsel.si_flags = 0;
    md->wsel.si_pid = 0;
    md->rxbytes = md->txbytes = 0;
    
    md->state = MD_OPEN;			/* port is now open, no other flags set */
    md->speed = 9600;				/* default to 9600 bps */

    disable_intr();				/* let nothing happen here */
    m_outb(md->iobase, com_lctl, CFCR_DLAB);
    m_outb(md->iobase, com_dlbl, combrd(md->clk, 9600) & 0xff);
    m_outb(md->iobase, com_dlbh, (u_int) combrd(md->clk, 9600) >> 8);
    m_outb(md->iobase, com_lctl, LCR_NORM);		/* = 8 bits, parity enabled, space parity */
    set_mcr(md, MD_RX);					/* set for receive */
    (void) m_inb(md->iobase, com_lsr);
    (void) m_inb(md->iobase, com_data);
    m_outb(md->iobase, com_ier, IER_ERXRDY | IER_ETXRDY);	/* enable interrupts */
    enable_intr();
    md->comphn = timeout(mdsiocomplain, md, hz/2);	/* run the whiner timeout */

    splx(s);
    debug("OK");
    
    return(0);
}

/*
** mdsioclose
**
** Disable interrupts and mark the port as closed.
*/

int
mdsioclose(dev_t dev, int flag __unused, int mode __unused, struct proc *p __unused) {
    struct mdsio_softc	*md;
    int			unit;
    int			s;
    int			rv;
    
    unit = minor(dev);
    md = devclass_get_softc(mdsio_devclass, unit);
    if (!(md->state & MD_OPEN))		/* XXX not open; what to return? */
	return(ENXIO);

    /* XXX should sleep here to wait for TX output to finish?? */
    s = spltty();

    while (txbuflen(md)) {		/* wait for buffer to empty */
	debug("sleep waiting for tx drain (%d)",txbuflen(md));
	md->state |= MD_DRAINSLEEP;	/* waiting for write activity */
    	rv = tsleep((caddr_t)md, TTIPRI | PCATCH, "mdsiowr", 0);
	md->state &= ~MD_DRAINSLEEP;	/* yawn */
	if (rv)				/* non-zero means interrupted*/
	    break;
    }

    m_outb(md->iobase, com_ier, 0);		/* disable interrupts */
    set_mcr(md, MD_RX);				/* disable drive */
    md->state = 0;				/* port closed */
    untimeout(mdsiocomplain, md, md->comphn);	/* kill the whiner timeout */
    splx(s);
    debug("OK");

    return(0);
}


/*
** mdsioread
**
** Suck data out of the ringbuffer and give it to the caller.
**
** We could be bastards and only copy the next contiguous fragment,
** but in the long run that's probably about as inefficient as
** using a bouncebuffer, so we do.
*/

int
mdsioread(dev_t dev, struct uio *uio, int flag) {
    struct mdsio_softc	*md;
    int			unit;
    int			s;
    int			rv = 0;
    u_int		hmuch, frag;
   
    unit = minor(dev);

    md = devclass_get_softc(mdsio_devclass, unit);

    if (uio->uio_resid % sizeof(u_short))	/* reads must be a multiple of short size */
	return(EINVAL);

    s = spltty();
    if (rxbuflen(md) == 0) {		/* no data ready, must sleep */
	if (flag & IO_NDELAY) {		/* can't sleep, return to caller */
	    splx(s);
	    return(EAGAIN);
	}
	md->state |= MD_READSLEEP;	/* make it so */
	splx(s);
	rv = tsleep((caddr_t)md, TTIPRI | PCATCH, "mdsiord", 0);
	md->state &= ~MD_READSLEEP;	/* yawn */
    }

    if (rv) {				/* nonzero means interrupted? */
	splx(s);
	return(EINTR);
    }
    
    /*
    ** There's some data to be read; hand it back.
    */

    hmuch = min(rxbuflen(md), uio->uio_resid/sizeof(u_short));	/* most we can give/is wanted */
    frag = min(hmuch, MDSIO_RXBUF - md->rxtail);		/* first slab */
    rv = uiomove((caddr_t)(md->rxbuf + md->rxtail), 		/* move first fragment */
		 frag * sizeof(u_short), uio);
    if (!rv && (frag < hmuch))					/* ok, but need more? */
	rv = uiomove((caddr_t)md->rxbuf,			/* move second fragment */
		     (hmuch - frag) * sizeof(u_short), uio);
    if (rv) {
	debug("uiomove returned %d",rv);
    } else {
	md->rxtail = (md->rxtail + hmuch) % MDSIO_RXBUF;	/* read OK, move the tail */
	debug("OK");
    }
    splx(s);
    return(rv);
}

/*
** mdsiowrite
**
** Stuff outgoing data into the transmit ringbuffer, kick things off
** appropriately to make sure it goes.
*/

int
mdsiowrite(dev_t dev, struct uio *uio, int flag) {
    struct mdsio_softc	*md;
    int			unit;
    int			s;
    int			rv;
    u_int		hmuch,frag;
 
    unit = minor(dev);

    md = devclass_get_softc(mdsio_devclass, unit);

    if ((uio->uio_resid % sizeof(u_short)) ||			/* writes must be multiple of u_short size */
	(uio->uio_resid > (int)(MDSIO_TXBUF*sizeof(u_short)))) {/* and must be smaller than the buffer size */
	debug("attempt to write %d bytes rejected",uio->uio_resid);
	return(EINVAL);
    }
    hmuch = uio->uio_resid / sizeof(short);	/* size of write operation (shorts) */
    s = spltty();				/* tail/head updates maybe not atomic? */
    while (hmuch > txbuffree(md)) {		/* wait for enough buffer space */
	if (flag & IO_NDELAY) {			/* can't sleep */
	    splx(s);
	    return(EAGAIN);
	}
	debug("sleep waiting for write space (have %d, want %d)",txbuffree(md),hmuch);
	md->state |= MD_WRITESLEEP;		/* waiting for a write to make room */
    	rv = tsleep((caddr_t)md, TTIPRI | PCATCH, "mdsiowr", 0);
	md->state &= ~MD_WRITESLEEP;		/* yawn */
	if (rv)	{				/* nonzero means interrupted? */
	    splx(s);
	    return(EINTR);
	}
    }
    splx(s);
    
    frag = min(hmuch, MDSIO_TXBUF - md->txhead);		/* how much to copy first time */
    rv = uiomove((caddr_t)(md->txbuf + md->txhead),		/* move first fragment */
		 frag * sizeof(u_short), uio);
    if (!rv && (frag < hmuch))
	rv = uiomove((caddr_t)md->txbuf,			/* move second fragment */
		     (hmuch - frag) * sizeof(u_short), uio);
    if (rv) {
	debug("uiomove returned %d",rv);
	return(rv);
    }
    md->txhead = (md->txhead + hmuch) % MDSIO_TXBUF;		/* update head pointer */

    disable_intr();
    set_mcr(md, MD_TX);						/* go to Tx mode regardless */
    debug("Faking interrupt for Tx kickoff");
    mdsiointr(md);						/* simulate interrupt to start xmit */
    enable_intr();

    md->state |= MD_XMITTING;					/* set before checking for timeout */
    md->state &= ~MD_TAILDRAIN;					/* clear taildrain flag */
    if (!(md->state & MD_TIMER)) {				/* no timeout looking after this port? */
	md->tmhn = timeout(mdsiotimeout, (void *)md, hz/MDSIO_TIMEOUT);	/* register timeout for this port */
	md->state |= MD_TIMER;
    }
    debug("OK");

    return(0);							/* our work here is done */
}


/*
** mdsiointr
**
** Interrupt handler
**
** Data received while transmitting is discarded, data received with 
** parity errors (address markers) has bit 8 set in the buffer.  Data
** received with framing errors are discarded (can be caused by another
** device coming onto/going off the bus, or noise).
**
** XXX for RS-422 use, the rx/tx interlock should be removed.
*/
    
void
mdsiointr(void *arg) {
    struct mdsio_softc	*md = (struct mdsio_softc *)arg;
    u_char		line_status;
    u_short		rdata;
    int			readsome = 0, wrotesome = 0;
    
    debug("called");
    if (!(md->state & MD_OPEN)) {			/* shouldn't interrupt while closed! */
	error("interrupt while closed");
	return;
    }
    md->ints++;
    
    for (;;) {						/* loop until no interrupting conditions */
	line_status = m_inb(md->iobase, com_lsr);

	while(line_status & LSR_RXRDY) {		/* loop reading first */
	    devstat_start_transaction(md->md_ds);
	    if (line_status & LSR_OE) {			/* overrun? */
		md->eflag |= MD_FIFOOF;
		md->errors++;
	    }
	    if (line_status & LSR_FE) {			/* framing error? */
		md->eflag |= MD_FERROR;			/* probably worth mentioning */
		md->errors++;
	    }
	    if (line_status & LSR_RXRDY) {
		rdata = m_inb(md->iobase, com_data);
		if (line_status & LSR_PE)			/* address marker */
		    rdata |= MD_ADDRMARK;
		if (rxheadnext(md) == md->rxtail) {		/* buffer full 8( */
		    md->eflag |= MD_IBUFOF;			/* mark for later whining */
		    md->errors++;
		} else {
			if ((md->state & MD_XMITTING) &&	/* transmitting */ 
			    (md->options & MD_2WIRE)) {		/* in 2-wire mode */
			    md->eflag |= MD_RXWHILETX;		/* shouldn't happen ?? */
			    md->errors++;
			} else {
			    md->rxbuf[md->rxhead] = rdata;	/* save in buffer */
			    md->rxhead = rxheadnext(md);
			    readsome++;				/* flag for later wakeup of reader */
			    devstat_end_transaction(md->md_ds, readsome, DEVSTAT_TAG_NONE,
						    DEVSTAT_READ);
			}
		}
	    }
	    line_status = m_inb(md->iobase, com_lsr);
	    md->rxbytes += readsome;
	}						/* end read loop */
				
	if (readsome) {					/* anything in? */
	    if (md->state & MD_READSLEEP) {		/* sleeping on read? */
		md->state &= ~MD_READSLEEP;		/* not anymore */
		wakeup((caddr_t)md);			/* kick */
	    } else
		selwakeup(&md->rsel);
	}

	if ((line_status & LSR_TXRDY) && 		/* UART ready for tx */
	    !(md->state & (MD_PREADDR|MD_POSTADDR)) && 	/* not held up on an address marker */
	    (md->state & MD_XMITTING)) {		/* we are actually doing something */

	    if (!txbuflen(md)) {			/* nothing to send */
		md->state |= MD_TAILDRAIN;		/* tell timeout to disable Tx later */
		goto tx_done;
	    }
	    if (md->txbuf[md->txtail] & MD_ADDRMARK) {	/* next byte is an address marker? */
		md->state |= MD_PREADDR;		/* tell the timeout code */
		goto tx_done;
	    }
	    devstat_start_transaction(md->md_ds);
	    m_outb(md->iobase, com_data,md->txbuf[md->txtail]&0xff);	/* write the low byte (data) */
	    devstat_end_transaction(md->md_ds, 1, DEVSTAT_TAG_NONE,
				DEVSTAT_WRITE);
	    md->txtail = txtailnext(md);
	    wrotesome++;				/* flag for later wakeup of writer */
	    md->txbytes += wrotesome;
	}
    tx_done:
	if (wrotesome) {				/* anything out? */
	    if (md->state & MD_WRITESLEEP) {		/* sleeping on write? */
		md->state &= ~MD_WRITESLEEP;		/* not anymore */
		wakeup((caddr_t)md);			/* kick */
	    } else
 		selwakeup(&md->wsel);
	}
	if ((m_inb(md->iobase, com_iir) & 0xf) == 1)		/* no activity pending */
	    return;
    }
}

/*
** mdsiotimeout
**
** Runs while transmitting, and handles waiting for the TX shift register
** to empty before and after address markers. 
**
** XXX may reqire spltty() protection?
*/

static void
mdsiotimeout(void *what) {
    struct mdsio_softc	*md;
    u_char		line_status;

    md = (struct mdsio_softc *)what;		/* this is what we work with */
    debug("called");

    if (md->state & MD_XMITTING)		/* timeout still wanted? */
	md->tmhn = timeout(mdsiotimeout, (void *)md, hz/MDSIO_TIMEOUT); /* register for another timeout */
    else {
	md->state &= ~MD_TIMER;			/* mark the timeout as not running */
	return;					/* timeout off now */
    }

    line_status = m_inb(md->iobase, com_lsr);	/* get the line status */
    if (!(line_status & LSR_TSRE))		/* shift register empty? */
	return;					/* nothing to do here, get out fast */
	
    if (md->state & MD_PREADDR) {		/* waiting for drain before address byte */
	m_outb(md->iobase, com_lctl, LCR_ADDR);	/* = 8 bits, parity enabled, mark parity */
#ifdef MDSIO_TXPREAMBLE
	/*
	 * In order to work with the combination of unknown network termination,
	 * PC interface cards with automatic drive control and receiving UARTs
	 * that require a mark/space transition before detecting a start bit,
	 * we transmit a preamble address byte of zero.  This is a hack.
	 */
	m_outb(md->iobase, com_data,0);		/* send a 0 preamble first, goes straight to TSR */
#endif
	m_outb(md->iobase, com_data, md->txbuf[md->txtail]&0xff);	/* write the low byte (data) */
	md->txtail = txtailnext(md);		/* indicate that the byte's gone */
	md->state |= MD_POSTADDR;		/* now waiting for it to clear */
	md->state &= ~MD_PREADDR;
	if (md->state & MD_WRITESLEEP) {	/* sleeping on write? */
	    md->state &= ~MD_WRITESLEEP;	/* not anymore */
	    wakeup((caddr_t)md);		/* kick */
	} else
	    selwakeup(&md->wsel);
	
	return;
    }
    if (md->state & MD_POSTADDR) {		/* waiting for address byte to drain */
	m_outb(md->iobase, com_lctl, LCR_NORM);	/* = 8 bits, parity enabled, space parity */
	md->state &= ~MD_POSTADDR;		/* not waiting anymore */
	disable_intr();
	debug("Faking intr for tx in mdsiotimeout");
	mdsiointr(md);				/* fake an interrupt for this unit to get tx going */
	enable_intr();
	return;
    }
    if (md->state & MD_TAILDRAIN) {		/* waiting for last byte to go? */
	set_mcr(md, MD_RX);			/* safe to let the bus go now */
	md->state &= ~(MD_TAILDRAIN|MD_XMITTING);	/* clear both states */
	if (md->state & MD_DRAINSLEEP) {	/* sleeping on drain? */
	    md->state &= ~MD_DRAINSLEEP;	/* not anymore */
	    wakeup((caddr_t)md);		/* kick */
	}
	return;
    }
    
    disable_intr();
    debug("Faking intr for tx loss");
    mdsiointr(md);				/* if we get here, a transmit interrupt has been lost */
    enable_intr();

    debug("OK");
}

/*
** mdsioioctl
**
** Ye Olde Ioctl handler.
**
** Not much you can do with one of these ports apart from set the baudrate.
*/

int
mdsioioctl(dev_t dev, u_long cmd, caddr_t data, int flag, struct proc *p) {
    struct mdsio_softc	*md;
    int			unit;
    int			s;
    int			*val = (int *)data;
    u_char		lcr_image;
		
    unit = minor(dev);
    md = devclass_get_softc(mdsio_devclass, unit);
    
    switch(cmd) {
    case MDSIOCSETSPEED:
	switch(*val)
	{
	case 1200:		/* reasonable values */
	case 2400:
	case 4800:
	case 9600:
	case 19200:
	case 38400:
	case 57600:
	case 115200:
	    break;
	    
	default:		/* unreasonable values */
	    return(EINVAL);
	}

	s = spltty();		/* would be Very Bad to have an interrupt */
	lcr_image = m_inb(md->iobase, com_lctl);
	m_outb(md->iobase, com_lctl, CFCR_DLAB);
	m_outb(md->iobase, com_dlbl, combrd(md->clk, *val) & 0xff);
	m_outb(md->iobase, com_dlbh, (combrd(md->clk, *val) >> 8) && 0xff);
	m_outb(md->iobase, com_lctl, lcr_image);
	debug("BL - 0x%02x", combrd(md->clk, *val) & 0xff);
	debug("BH - 0x%02x", (combrd(md->clk, *val) >> 8) & 0xff);
	splx(s);
	return(0);
	
    case MDSIOCGETSPEED:
	*val = md->speed;
	return(0);

    case FIONBIO:		/* ignore these */
    case FIOASYNC:
	return(0);
	
    }
    error("got unknown cmd 0x%lx",cmd);
    return(ENOTTY);
}

/*
** mdsiopoll
**
** Determine whether there's data available to read/ room to write
*/

int 
mdsiopoll(dev_t dev, int events, struct proc *p) {
    struct mdsio_softc	*md;
    int			unit;
    int			revents = 0;
    int			s;
    
    unit = minor(dev);
    md = devclass_get_softc(mdsio_devclass, unit);

    s = spltty();			/* lock buffer stats */
    if (events & (POLLIN | POLLRDNORM)) {
	if (!rxbuflen(md)) {		/* data available to read */
	    debug("called for read, no data");
	    selrecord(p, &md->rsel);
	} else {
	    debug("called for read, data");
	    revents |= events & (POLLIN | POLLRDNORM);
	}
    }
    
    if (events & (POLLOUT | POLLWRNORM)) {
	if (!txbuffree(md)) {		/* room to write now */
	    debug("called for write, no room");
	    selrecord(p, &md->wsel);
	} else {
	    debug("called for write, room");
	    revents |= events & (POLLOUT | POLLWRNORM);
	}
    }
    
    splx(s);
    return(revents);
}

/*
** mdsiocomplain
**
** periodic error reporter, runs while a port is open
*/

static void
mdsiocomplain(void *what) {
    struct mdsio_softc	*md = what;
    int			i,j,s;
    
    timeout(mdsiocomplain, md, hz/2);	/* every half a second */
    
    s = spltty();
    i = md->eflag;			/* get error flag safely */
    md->eflag = 0;
    j = md->errors;
    splx(s);

    if (i & MD_FIFOOF)
	error("FIFO overflow");
    if (i & MD_IBUFOF)
	error("interrupt-level buffer overflow");
    if (i & MD_FERROR)
	error("framing error");
    if (i & MD_RXWHILETX)
	error("received data while transmitting");
    if (i)
	error("total errors : %d",j);
}

/*
** combrd
**
** Work out the clock divisor needed for a given baud rate
*/
static int
combrd(int clk, int baud) {
    return (clk / (16 * baud));
}

/*
** sysctl_hw_mdsio_txcontrol
**
** Get/set the device flags which will change the TX control method
*/
static int
sysctl_hw_mdsio_txcontrol(SYSCTL_HANDLER_ARGS) {
/* struct sysctl_oid *oidp, void *arg1, int arg2, struct sysctl_req *req */
    device_t	dev = (device_t)oidp->oid_arg1;
#ifdef MDSIODEBUG
    struct mdsio_softc	*md = device_get_softc(dev);
#endif
    int		error, value;

    if (req->newptr != NULL) {
	debug("Setting");
	
	error = SYSCTL_IN(req, &value, sizeof(int));
	if (error) {
	    debug("SYSCTL_IN failed %d", error);
	    return (error);
	}
	debug("sysctl called with 0x%02x", value);
	device_set_flags(dev, value);
	error = SYSCTL_OUT(req, &value, sizeof(int));
	if (error) {
	    debug("SYSCTL_OUT failed %d", error);
	    return(error);
	}
    
	mdsio_update_options(dev);
    } else {
	debug("Getting");
	value = device_get_flags(dev);
	error = SYSCTL_OUT(req, &value, sizeof(int));
	if (error) {
	    debug("SYSCTL_OUT failed %d", error);
	    return (error);
	}
    }
    
    return(0);
}

static void
mdsio_update_options(device_t dev) {
    struct mdsio_softc	*md = device_get_softc(dev);

    debug("Called with flags 0x%02x", device_get_flags(dev));
    if (md->options & MD_TXCTL)
	device_printf(dev, "%d-wire mode,%s %s transmit control.\n",
	       (md->options & MD_2WIRE) ? 2:4,
	       TXCTRL_INVERT(dev) ? " inverted":"",
	       DTR_TXCONTROL(dev) ? "DTR":"RTS");
    else
	device_printf(dev, "%d-wire mode, automatic transmit control.\n",
	       (md->options & MD_2WIRE) ? 2:4);

    md->options = 0;
    if (MDSIO_2WIRE(dev))		/* 2-wire mode? */
	md->options |= MD_2WIRE;

    if (MDSIO_TXCONTROL(dev)) {	/* Tx/Rx control required? */
	u_char	ass = 0;
	
	md->options |= MD_TXCTL;
	md->tx_mcr = md->rx_mcr = 0;
	if (DTR_TXCONTROL(dev))
	    ass = MCR_DTR;
	if (RTS_TXCONTROL(dev))		/* RTS more common, assume in donkey case */
	    ass = MCR_RTS;
	if (TXCTRL_INVERT(dev)) {	/* flipped? */
	    md->tx_mcr = MCR_IENABLE;   /* line low for TX */
	    md->rx_mcr = ass | MCR_IENABLE;
	} else {
	    md->tx_mcr = ass | MCR_IENABLE; /* line high for TX */
	    md->rx_mcr = MCR_IENABLE;
	}
    } else {
	md->tx_mcr = md->rx_mcr = MCR_IENABLE;
    }

    set_mcr(md, MD_RX);			/* go immediately to receive mode */
}
